@@ -129,7 +129,6 @@ void ScreenCastStream::onStreamStateChanged(pw_stream_state old, pw_stream_state
             Q_EMIT ready(nodeId());
         }
         m_pendingFrame.stop();
-        m_pendingDamage = QRegion();
         m_pendingContents = Contents();
         m_source->pause();
         break;
@@ -291,6 +290,7 @@ void ScreenCastStream::onStreamAddBuffer(pw_buffer *pwBuffer)
                                                                        .modifiers = {m_videoFormat.modifier},
                                                                    })) {
             pwBuffer->user_data = dmabuf;
+            m_allBuffers.push_back(dmabuf);
             return;
         }
     }
@@ -302,6 +302,7 @@ void ScreenCastStream::onStreamAddBuffer(pw_buffer *pwBuffer)
                                                                      .software = true,
                                                                  })) {
             pwBuffer->user_data = memfd;
+            m_allBuffers.push_back(memfd);
             return;
         }
     }
@@ -312,6 +313,7 @@ void ScreenCastStream::onStreamRemoveBuffer(pw_buffer *pwBuffer)
     if (ScreenCastBuffer *buffer = static_cast<ScreenCastBuffer *>(pwBuffer->user_data)) {
         delete buffer;
         pwBuffer->user_data = nullptr;
+        m_allBuffers.removeOne(buffer);
     }
 
     m_dequeuedBuffers.removeOne(pwBuffer);
@@ -323,8 +325,8 @@ ScreenCastStream::ScreenCastStream(ScreenCastSource *source, std::shared_ptr<Pip
     , m_source(source)
     , m_resolution(source->textureSize())
 {
-    connect(source, &ScreenCastSource::frame, this, [this](const QRegion &damage) {
-        scheduleRecord(damage, Content::Video);
+    connect(source, &ScreenCastSource::frame, this, [this]() {
+        scheduleRecord(Content::Video);
     });
     connect(source, &ScreenCastSource::closed, this, &ScreenCastStream::close);
 
@@ -348,8 +350,7 @@ ScreenCastStream::ScreenCastStream(ScreenCastSource *source, std::shared_ptr<Pip
 
     m_pendingFrame.setSingleShot(true);
     connect(&m_pendingFrame, &QTimer::timeout, this, [this] {
-        record(m_pendingDamage, m_pendingContents);
-        m_pendingDamage = QRegion();
+        record(m_pendingContents);
         m_pendingContents = Contents();
     });
 }
@@ -446,7 +447,7 @@ bool ScreenCastStream::createStream()
     case ScreencastV1Interface::Metadata:
         m_cursor.changedConnection = connect(Cursors::self(), &Cursors::currentCursorChanged, this, &ScreenCastStream::invalidateCursor);
         m_cursor.positionChangedConnection = connect(Cursors::self(), &Cursors::positionChanged, this, [this] {
-            scheduleRecord({}, Content::Cursor);
+            scheduleRecord(Content::Cursor);
         });
         break;
     }
@@ -474,12 +475,12 @@ void ScreenCastStream::close()
     disconnect(m_cursor.positionChangedConnection);
     m_cursor.positionChangedConnection = {};
 
-    m_source->pause();
+    m_source.reset();
 
     Q_EMIT closed();
 }
 
-void ScreenCastStream::scheduleRecord(const QRegion &damage, Contents contents)
+void ScreenCastStream::scheduleRecord(Contents contents)
 {
     Q_ASSERT(!m_closed);
 
@@ -498,25 +499,21 @@ void ScreenCastStream::scheduleRecord(const QRegion &damage, Contents contents)
         }
     }
 
+    m_pendingContents |= contents;
+
     if (m_pendingFrame.isActive()) {
-        m_pendingDamage += damage;
-        m_pendingContents |= contents;
         return;
     }
-
+    std::chrono::milliseconds waitInterval{0};
     if (m_videoFormat.max_framerate.num != 0 && m_lastSent.has_value()) {
         const auto now = std::chrono::steady_clock::now();
         const auto frameInterval = std::chrono::milliseconds(1000 * m_videoFormat.max_framerate.denom / m_videoFormat.max_framerate.num);
         const auto lastSentAgo = std::chrono::duration_cast<std::chrono::milliseconds>(now - m_lastSent.value());
         if (lastSentAgo < frameInterval) {
-            m_pendingDamage += damage;
-            m_pendingContents |= contents;
-            m_pendingFrame.start(frameInterval - lastSentAgo);
-            return;
+            waitInterval = frameInterval - lastSentAgo;
         }
     }
-
-    record(damage, contents);
+    m_pendingFrame.start(waitInterval);
 }
 
 pw_buffer *ScreenCastStream::dequeueBuffer()
@@ -570,7 +567,7 @@ pw_buffer *ScreenCastStream::dequeueBuffer()
     return pwBuffer;
 }
 
-void ScreenCastStream::record(const QRegion &damage, Contents contents)
+void ScreenCastStream::record(Contents contents)
 {
     EglBackend *backend = qobject_cast<EglBackend *>(Compositor::self()->backend());
     if (!backend) {
@@ -588,11 +585,21 @@ void ScreenCastStream::record(const QRegion &damage, Contents contents)
     ScreenCastBuffer *buffer = static_cast<ScreenCastBuffer *>(pwBuffer->user_data);
 
     Contents effectiveContents = contents;
-    if (m_cursor.mode != ScreencastV1Interface::Hidden) {
-        effectiveContents.setFlag(Content::Cursor);
-        if (m_cursor.mode == ScreencastV1Interface::Embedded) {
-            effectiveContents.setFlag(Content::Video);
+    switch (m_cursor.mode) {
+    case ScreencastV1Interface::Hidden:
+        m_source->setRenderCursor(false);
+        break;
+    case ScreencastV1Interface::Metadata:
+        effectiveContents |= Content::Cursor;
+        m_source->setRenderCursor(false);
+        if (effectiveContents & Content::Cursor) {
+            addCursorMetadata(spa_buffer, Cursors::self()->currentCursor());
         }
+        break;
+    case ScreencastV1Interface::Embedded:
+        effectiveContents |= Content::Cursor | Content::Video;
+        m_source->setRenderCursor(m_source->includesCursor(Cursors::self()->currentCursor()));
+        break;
     }
 
     EglContext *context = backend->openglContext();
@@ -600,9 +607,11 @@ void ScreenCastStream::record(const QRegion &damage, Contents contents)
 
     spa_meta_sync_timeline *synctmeta = nullptr;
 
+    QRegion damage;
     if (effectiveContents & Content::Video) {
         if (auto memfd = dynamic_cast<MemFdScreenCastBuffer *>(buffer)) {
-            m_source->render(memfd->view.image());
+            damage = m_source->render(memfd->view.image(), m_damageJournal.accumulate(memfd->m_age, infiniteRegion()));
+            bumpBufferAge(memfd);
         } else if (auto dmabuf = dynamic_cast<DmaBufScreenCastBuffer *>(buffer)) {
             if (dmabuf->synctimeline) {
                 synctmeta = static_cast<spa_meta_sync_timeline *>(spa_buffer_find_meta_data(spa_buffer,
@@ -615,23 +624,10 @@ void ScreenCastStream::record(const QRegion &damage, Contents contents)
                 }
             }
 
-            m_source->render(dmabuf->framebuffer.get());
-        }
-    }
-
-    QRegion effectiveDamage = damage;
-    if (effectiveContents & Content::Cursor) {
-        Cursor *cursor = Cursors::self()->currentCursor();
-        switch (m_cursor.mode) {
-        case ScreencastV1Interface::Hidden:
-            break;
-        case ScreencastV1Interface::Embedded:
-            effectiveDamage += addCursorEmbedded(buffer, cursor);
-            break;
-        case ScreencastV1Interface::Metadata:
-            addCursorMetadata(spa_buffer, cursor);
-            break;
+            damage = m_source->render(dmabuf->framebuffer.get(), m_damageJournal.accumulate(dmabuf->m_age, infiniteRegion()));
+            bumpBufferAge(dmabuf);
         }
+        m_damageJournal.add(damage);
     }
 
     if (spa_data[0].type == SPA_DATA_DmaBuf) {
@@ -653,7 +649,7 @@ void ScreenCastStream::record(const QRegion &damage, Contents contents)
         }
     }
 
-    addDamage(spa_buffer, effectiveDamage);
+    addDamage(spa_buffer, damage);
     addHeader(spa_buffer);
 
     if (effectiveContents & Content::Video) {
@@ -669,6 +665,17 @@ void ScreenCastStream::record(const QRegion &damage, Contents contents)
     resize(m_source->textureSize());
 }
 
+void ScreenCastStream::bumpBufferAge(ScreenCastBuffer *renderedBuffer)
+{
+    for (ScreenCastBuffer *buffer : std::as_const(m_allBuffers)) {
+        if (buffer == renderedBuffer) {
+            buffer->m_age = 1;
+        } else if (buffer->m_age > 0) {
+            buffer->m_age++;
+        }
+    }
+}
+
 void ScreenCastStream::resize(const QSize &resolution)
 {
     if (m_resolution == resolution) {
@@ -859,59 +866,6 @@ void ScreenCastStream::addCursorMetadata(spa_buffer *spaBuffer, Cursor *cursor)
     }
 }
 
-QRegion ScreenCastStream::addCursorEmbedded(ScreenCastBuffer *buffer, Cursor *cursor)
-{
-    if (!m_source->includesCursor(cursor)) {
-        const QRegion damage = m_cursor.lastRect.toAlignedRect();
-        m_cursor.visible = false;
-        m_cursor.lastRect = QRectF();
-        return damage;
-    }
-
-    const QRectF cursorRect = scaledRect(m_source->mapFromGlobal(cursor->geometry()), m_source->devicePixelRatio());
-    if (auto memfd = dynamic_cast<MemFdScreenCastBuffer *>(buffer)) {
-        QPainter painter(memfd->view.image());
-        const PlatformCursorImage cursorImage = kwinApp()->cursorImage();
-        painter.drawImage(cursorRect, cursorImage.image());
-    } else if (auto dmabuf = dynamic_cast<DmaBufScreenCastBuffer *>(buffer)) {
-        if (m_cursor.invalid) {
-            m_cursor.invalid = false;
-            const PlatformCursorImage cursorImage = kwinApp()->cursorImage();
-            if (cursorImage.isNull()) {
-                m_cursor.texture = nullptr;
-            } else {
-                m_cursor.texture = GLTexture::upload(cursorImage.image());
-            }
-        }
-
-        if (m_cursor.texture) {
-            GLFramebuffer::pushFramebuffer(dmabuf->framebuffer.get());
-
-            auto shader = ShaderManager::instance()->pushShader(ShaderTrait::MapTexture);
-
-            QMatrix4x4 mvp;
-            mvp.scale(1, -1);
-            mvp.ortho(QRectF(QPointF(0, 0), dmabuf->texture->size()));
-            mvp.translate(cursorRect.x(), cursorRect.y());
-            shader->setUniform(GLShader::Mat4Uniform::ModelViewProjectionMatrix, mvp);
-
-            glEnable(GL_BLEND);
-            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
-            m_cursor.texture->render(cursorRect.size());
-            glDisable(GL_BLEND);
-
-            ShaderManager::instance()->popShader();
-            GLFramebuffer::popFramebuffer();
-        }
-    }
-
-    const QRegion damage = QRegion{m_cursor.lastRect.toAlignedRect()} | cursorRect.toAlignedRect();
-    m_cursor.visible = true;
-    m_cursor.lastRect = cursorRect;
-
-    return damage;
-}
-
 void ScreenCastStream::setCursorMode(ScreencastV1Interface::CursorMode mode)
 {
     m_cursor.mode = mode;
